# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/8/16 14:17
Desc:
'''
# glm中的pvalue值??

import statsmodels.api as sm
import os
import numpy as np
import pandas as pd
from itertools import repeat
from itertools import combinations
from multiprocessing import Pool
import math
from sklearn.model_selection import train_test_split
from .util import _Tool,SCORERS
import logging


class Regression():
    def __init__(self, X:pd.DataFrame, Y, init_in_vars=[], model=None, sample_weight = None, max_related_level=1, iter_num=20,
                 is_train_test_split=True, test_size=0.4, split_num=10, kw_measure_args={},
                 measure = 'ks', max_pvalue_limit=0.05, max_vif_limit=10, max_corr_limit=0.6,seed=666,
                n_core=None, init_dataset=[],logger_info=True, logger_file=None, observation_model='normal'):
        # fit_weight = None, measure_weight = None, kw_measure_args = None,
        # kw_algorithm_class_args = None


        '''
        :param X: 数据特征集合,pd.DataFrame
        :param Y: 数据标签，一维list
        :param model: 评估所使用的模型
        :param iter_num: 最大迭代次数,预估筛选出多少个特征
        :param measure: 评估指标
        :param max_pvalue_limit: 错误拒绝H0的概率很低，表示H0本身(两者无关联)就是错误的。两者两者有关联，这个独立变量重要。
        :param max_vif_limit: 方差膨胀系数，是对多重共线性严重程度的一种度量。它表示回归系数估计量的方差与假设自变量间不线性相关时方差
                相比的比值。一般>5或者>10表示严重多重共线
        :param max_corr_limit: 皮尔逊相关系数。值越大表示越相关
        :param n_core: 进程数
        :param logger_file: 日志地址
        :param init_dataset: 测验数据集=[[X_train, X_valid, Y_train, Y_valid],...]
        '''
        if model is None:
            self.model = self._regression
        else:
            self.model=model
        self.sample_weight=sample_weight
        self.measure = measure
        self.max_related_level=max_related_level
        self.max_pvalue_limit = max_pvalue_limit
        self.max_vif_limit = max_vif_limit
        self.max_corr_limit = max_corr_limit
        self.is_trte_split=is_train_test_split
        self.test_size=test_size
        self.split_num=split_num
        self.kw_measure_args = kw_measure_args
        self.iter_num = iter_num
        self.logger_info = logger_info
        self.logger_file=logger_file
        self.seed=seed
        self.dataset=init_dataset
        self.in_vars=init_in_vars # 初始默认输入特征
        self.observation_model=observation_model # 'all'代表全变量测试，观测所有特征的效果趋势
        self.is_check_overfit=False

        if n_core is None:
            self.n_core = os.cpu_count() - 1
        elif n_core >= 1:
            self.n_core = n_core
        else:
            self.n_core = math.ceil(os.cpu_count() * n_core)

        self._check_input_data(X,Y)

    def _check_input_data(self,X,Y):
        if self.dataset==[]:
            dataset = []
            if self.is_trte_split:
                seed_state = np.random.RandomState(seed=self.seed)
                seed_set = [int(_ * 10000) for _ in seed_state.rand(self.split_num)]
                for seed_i in seed_set:
                    X_train, X_valid, Y_train, Y_valid = train_test_split(X, Y, test_size=0.4,random_state=seed_i)
                    dataset.append([X_train, X_valid, Y_train, Y_valid])
            else:
                X_train, X_valid, Y_train, Y_valid = X, Y, X, Y
                dataset.append([X_train, X_valid, Y_train, Y_valid])
            self.dataset=dataset
        assert len(self.dataset[0])==4
        self.columns=dataset[0][0].columns

    def _check(self, in_vars, current_perf):

        # 多次抽样
        pvalue_list=[]
        perf_list=[]
        for i in self.dataset:
            X_train, X_valid, Y_train, Y_valid =i
            X_train, X_valid=X_train[in_vars], X_valid[in_vars]
            clf = self.model(X_train,Y_train)
            Y_valid_pre = pd.Series(clf.predict(sm.add_constant(X_valid)), index=Y_valid.index, name='score')
            ipvalue = clf.pvalues ## 1.待修改
            pvalue_list.append(ipvalue)
            iperf = SCORERS[self.measure](Y_valid, Y_valid_pre, **self.kw_measure_args) ## 2.评价指标
            perf_list.append(iperf)
        df_pvalues =pd.concat(pvalue_list,axis=1).mean(1)

        perfs_mean = np.mean(perf_list)
        perfs_max = np.max(perf_list)
        perfs = perfs_max if self.is_check_overfit else perfs_mean

        X = pd.concat([X_train,X_valid],axis=0)
        if X.shape[1] < 2:
            corr_max = 0
            vif = 0
        else:
            vif = _Tool.vif(X).iloc[0, 1]
            df_corr = X.corr()
            t = np.arange(df_corr.shape[1])
            df_corr.values[t, t] = np.nan
            corr_max = df_corr.applymap(abs).max().max()

        check_pvalue = (df_pvalues < self.max_pvalue_limit).all()
        check_perf = (perfs > current_perf) and (perfs_mean>0.95*current_perf)
        check_vif = vif < self.max_vif_limit
        check_corr = corr_max < self.max_corr_limit

        return check_pvalue, check_perf, perfs, check_vif, vif, check_corr, corr_max

    def _add_var(self, args):
        col = args[0]
        in_vars, current_perf = args[1]
        add_rm_var = (None, None, current_perf)
        if isinstance(col,list):
            tmp_cols = []
            tmp_cols.extend(col)
        else:
            tmp_cols = [col]
        tmp_cols.extend(in_vars)

        check_pvalue, check_perf, perf, check_vif, vif, check_corr, corr_max = self._check(tmp_cols,current_perf)
        if check_perf:
            if check_pvalue and check_vif and check_corr:
                add_rm_var = (col, None, perf)
            else:
                if len(in_vars) > 0:
                    rm_var_arr = map(self._rm_var, zip(in_vars, repeat((tmp_cols, current_perf))))
                    rm_var, perf = sorted(rm_var_arr, key=lambda x: x[1])[-1]
                    if rm_var:
                        add_rm_var = (col, rm_var, perf)
        return add_rm_var

    def _rm_var(self, args):
        col = args[0]
        in_vars, current_perf = args[1]
        rm_var = (None, current_perf)
        X_col = list(set(in_vars)-set([col]))
        check_pvalue, check_perf, perf, check_vif, vif, check_corr, corr_max = self._check(X_col, current_perf)
        check_pass = (check_pvalue and check_vif and check_corr and check_perf)
        if check_pass:
            rm_var = (col, perf)
        return rm_var

    def _del_reason(self, args):
        col = args[0]
        in_vars, current_perf = args[1]
        tmp_cols = [col]
        tmp_cols.extend(in_vars)
        check_pvalue, check_perf, perf, check_vif, vif, check_corr, corr_max = self._check(tmp_cols,current_perf)

        reasons = []
        if not check_perf:
            reasons.append('模型性能=%f,小于等于最终模型的性能=%f' % (perf, current_perf))
        if not check_vif:
            reasons.append('最大VIF=%f,大于设置的阈值=%f' % (vif, self.max_vif_limit))
        if not check_corr:
            reasons.append('最大相关系数=%f,大于设置的阈值=%f' % (corr_max, self.max_corr_limit))
        if not check_pvalue:
            reasons.append('有些系数不显著，P_VALUE大于设置的阈值=%f' % (self.max_pvalue_limit))
        return (col, reasons)

    def fit(self):
        '''
        :return:
            in_vars : list,所有的入模变量列表，列表中的顺序即为加入时的顺序
            out_vars : dict,剔除的变量列表及删除原因，其结构为:{'变量名称':[删除原因]}
        '''
        iter_idx = 0
        in_vars = self.in_vars
        current_perf = -np.inf
        logger_ch, fh_ch = _Tool.make_logger(self.logger_info,self.logger_file)
        attempt_over=False
        attempt_str=''
        while (True):
            iter_idx += 1
            if iter_idx > self.iter_num:
                break
            if logger_ch:
                logger_ch.info('****************迭代轮数：%d********************' %iter_idx)

            out_vars = list(self.columns[~self.columns.isin(in_vars)])
            if len(out_vars) == 0:
                if logger_ch:
                    logger_ch.info('变量全部进入模型，建模结束')
                break
            ## out_vars 两级相关+三级
            if self.max_related_level>1:
                for i in range(2,self.max_related_level+1):
                    var_tmp=[list(i) for i in combinations(out_vars, i) if self._check_corr(i)]
                    out_vars.extend(var_tmp)
            with Pool(self.n_core) as pool:
                result = pool.map_async(self._add_var, zip(out_vars, repeat((in_vars, current_perf))))
                add_rm_var_arr = result.get()
            add_var, rm_var_0, perf = sorted(add_rm_var_arr, key=lambda x: x[2])[-1]
            if add_var is not None:
                if isinstance(add_var,list):
                    in_vars.extend(add_var)
                else:
                    in_vars.append(add_var)
                current_perf = perf
                if rm_var_0:
                    in_vars.remove(rm_var_0)
            if len(in_vars) == 0:
                if logger_ch:
                    logger_ch.info('没有变量能够进入模型，建模结束')
                break
            with Pool(self.n_core) as pool:
                result = pool.map_async(self._rm_var, zip(in_vars, repeat((in_vars, current_perf))))
                rm_var_arr = result.get()
            rm_var, perf = sorted(rm_var_arr, key=lambda x: x[1])[-1]
            if rm_var is not None:
                in_vars.remove(rm_var)
                current_perf = perf
            if (add_var is None) and (rm_var is None):
                if not attempt_over:
                    attempt_over=True
                    attempt_str='(overfit)'
                    self.is_check_overfit=True
                    continue
                if logger_ch:
                    logger_ch.info('在此轮迭代中，在满足使用者所设置条件的前提下，已经不能通过增加或删除变量来进一步提升模型的指标，建模结束')
                if self.observation_model=='all':
                    self.fit_observation(iter_idx, in_vars, logger_ch)
                break
            if logger_ch:
                logger_ch.info('此轮迭代完成%s，当前入模变量为：%s。 当前模型性能%s为:%f' % (attempt_str, in_vars, self.measure, current_perf))
        out_vars = self.columns[~self.columns.isin(in_vars)]
        with Pool(self.n_core) as pool:
            result = pool.map_async(self._del_reason, zip(out_vars, repeat((in_vars, current_perf))))
            del_var_arr = result.get()
        del_vars = dict((col, reasons) for col, reasons in del_var_arr)
        if fh_ch:
            fh_ch.close()
            logger_ch.removeHandler(fh_ch)
        return in_vars, del_vars


    def fit_observation(self,iter_idx,_in_vars,logger_ch):
        in_vars = _in_vars.copy()
        if logger_ch:
            logger_ch.info('****************迭代轮数(全量观察)：%s********************' % in_vars)
        while (True):
            iter_idx += 1
            if iter_idx > self.iter_num:
                break
            out_vars = list(self.columns[~self.columns.isin(in_vars)])
            if len(out_vars) == 0:
                if logger_ch:
                    logger_ch.info('(全量观察)变量全部进入模型，建模结束')
                break
            with Pool(self.n_core) as pool:
                result = pool.map_async(self._add_observation_var, zip(out_vars, repeat((in_vars, -1))))
                add_rm_var_arr = result.get()
            add_var, rm_var_0, perf = sorted(add_rm_var_arr, key=lambda x: x[2])[-1]
            if add_var is not None:
                if isinstance(add_var,list):
                    in_vars.extend(add_var)
                else:
                    in_vars.append(add_var)
                current_perf = perf
            if logger_ch:
                logger_ch.info('此轮迭代完成(全量观察)%d，入模变量为：%s。 当前模型性能%s为:%f' % (iter_idx, in_vars, self.measure, current_perf))

    def _add_observation_var(self, args):
        col = args[0]
        in_vars, current_perf = args[1]
        add_rm_var = (None, None, current_perf)
        if isinstance(col,list):
            tmp_cols = []
            tmp_cols.extend(col)
        else:
            tmp_cols = [col]
        tmp_cols.extend(in_vars)

        check_pvalue, check_perf, perf, check_vif, vif, check_corr, corr_max = self._check(tmp_cols,current_perf)
        add_rm_var = (col, None, perf)
        return add_rm_var

    def _check_corr(self,col):
        for i in col:
            assert i in self.columns,'{} 不是有效字段'.format(i)
        data = pd.concat(self.dataset[0][:2],axis=0)
        df_corr = data[list(col)].corr().applymap(abs)
        t = np.arange(df_corr.shape[1])
        df_corr.values[t, t] = np.nan
        corr_max = df_corr.max().max()
        if corr_max>self.max_corr_limit:
            return False
        else:
            return True

    def _regression(self,X,Y,weights=None,**kwargs):
        if weights is None:
            weights = np.ones(len(X))
        reg = sm.WLS(Y, sm.add_constant(X), weights=weights, **kwargs)
        clf = reg.fit()
        return clf


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/8/19 16:52
Desc:
'''
from pynovice.score_card import card_explanations
import numpy as np


class shap_values:
    pass

shap_values.base_values=10
shap_values.feature_names=['年龄','性别','身高']
shap_values.data=[30,'女',167]
shap_values.values=np.array([2,4,-3])
# pl.rcParams['font.sans-serif'] = ['SimSun']

# 1.单特征
# card_explanations.waterfall(base_value=10, shap_values=np.array([2,4,-3]), features=np.array([30,'女',167]),
#                                         feature_names=['年龄','性别','身高'])


# 2.单特征
# card_explanations.force(
#     base_value=10, shap_values=np.array([2,4,-3]), features=[30,'女',167],
#                                         feature_names=['年龄','性别','身高'],matplotlib=True)

# # 3.数据集特征贡献
card_explanations.beeswarm(base_values=561,shap_values=np.array([[2,4,-3]]), features=[[30,'女',167]],
                                        feature_names=['年龄','性别','身高'])


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-22 20:10
Desc:
    贝叶斯分bins （循环迭代找出最佳拟合度的拐点，拟合度(泊松分布)）
     泊松分布适合于描述单位时间内随机事件发生的次数。离散概率分布
    Bayesian Block implementation
    =============================
    Dynamic programming algorithm for finding the optimal adaptive-width histogram.
    Based on Scargle et al 2012 [1]_
    References
    ----------
    .. [1] http://adsabs.harvard.edu/abs/2012arXiv1207.5578S
    对于时间序列
    >>> t = [23,4,1,2,3,4,3,3,34,54,34,54,34,43,45,22,34,25]
    >>> bins = bayesian_blocks(t, fitness='events', p0=0.01)
    >>> print(bins)
    [-inf  3.5  inf]
'''

import numpy as np


class FitnessFunc:
    """Base class for fitness functions
    Each fitness function class has the following:
    - fitness(...) : compute fitness function.
       Arguments accepted by fitness must be among [T_k, N_k, a_k, b_k, c_k]
    - prior(N, Ntot) : compute prior on N given a total number of points Ntot
    """
    def __init__(self, p0=0.05, gamma=None):
        self.p0 = p0
        self.gamma = gamma

    def validate_input(self, t, x, sigma):
        """Check that input is valid"""
        pass

    def fitness(**kwargs):
        raise NotImplementedError()

    def prior(self, N, Ntot):
        if self.gamma is None:
            return self.p0_prior(N, Ntot)
        else:
            return self.gamma_prior(N, Ntot)

    def p0_prior(self, N, Ntot):
        # eq. 21 from Scargle 2012
        return 4 - np.log(73.53 * self.p0 * (N ** -0.478))

    def gamma_prior(self, N, Ntot):
        """Basic prior, parametrized by gamma (eq. 3 in Scargle 2012)"""
        if self.gamma == 1:
            return 0
        else:
            return (np.log(1 - self.gamma)
                    - np.log(1 - self.gamma ** (Ntot + 1))
                    + N * np.log(self.gamma))

    # the fitness_args property will return the list of arguments accepted by
    # the method fitness().  This allows more efficient computation below.
    @property
    def args(self):
        try:
            # Python 2
            return self.fitness.func_code.co_varnames[1:]
        except AttributeError:
            return self.fitness.__code__.co_varnames[1:]


class Events(FitnessFunc):
    """Fitness for binned or unbinned events
    Parameters
    ----------
    p0 : float
        False alarm probability, used to compute the prior on N
        (see eq. 21 of Scargle 2012).  Default prior is for p0 = 0.
    gamma : float or None
        If specified, then use this gamma to compute the general prior form,
        p ~ gamma^N.  If gamma is specified, p0 is ignored.
    """
    def fitness(self, N_k, T_k):
        # eq. 19 from Scargle 2012
        return N_k * (np.log(N_k) - np.log(T_k))

    def prior(self, N, Ntot):
        if self.gamma is not None:
            return self.gamma_prior(N, Ntot)
        else:
            # eq. 21 from Scargle 2012
            return 4 - np.log(73.53 * self.p0 * (N ** -0.478))


class RegularEvents(FitnessFunc):
    """Fitness for regular events
    This is for data which has a fundamental "tick" length, so that all
    measured values are multiples of this tick length.  In each tick, there
    are either zero or one counts.
    Parameters
    ----------
    dt : float
        tick rate for data
    gamma : float
        specifies the prior on the number of bins: p ~ gamma^N
    """
    def __init__(self, dt, p0=0.05, gamma=None):
        self.dt = dt
        self.p0 = p0
        self.gamma = gamma

    def validate_input(self, t, x, sigma):
        unique_x = np.unique(x)
        if list(unique_x) not in ([0], [1], [0, 1]):
            raise ValueError("Regular events must have only 0 and 1 in x")

    def fitness(self, T_k, N_k):
        # Eq. 75 of Scargle 2012
        M_k = T_k / self.dt
        N_over_M = N_k * 1. / M_k

        eps = 1E-8
        if np.any(N_over_M > 1 + eps):
            import warnings
            warnings.warn('regular events: N/M > 1.  '
                          'Is the time step correct?')

        one_m_NM = 1 - N_over_M
        N_over_M[N_over_M <= 0] = 1
        one_m_NM[one_m_NM <= 0] = 1

        return N_k * np.log(N_over_M) + (M_k - N_k) * np.log(one_m_NM)

class PointMeasures(FitnessFunc):
    """Fitness for point measures
    Parameters
    ----------
    gamma : float
        specifies the prior on the number of bins: p ~ gamma^N
        if gamma is not specified, then a prior based on simulations
        will be used (see sec 3.3 of Scargle 2012)
    """
    def __init__(self, p0=None, gamma=None):
        self.p0 = p0
        self.gamma = gamma

    def fitness(self, a_k, b_k):
        # eq. 41 from Scargle 2012
        return (b_k * b_k) / (4 * a_k)

    def prior(self, N, Ntot):
        if self.gamma is not None:
            return self.gamma_prior(N, Ntot)
        elif self.p0 is not None:
            return self.p0_prior(N, Ntot)
        else:
            # eq. at end of sec 3.3 in Scargle 2012
            return 1.32 + 0.577 * np.log10(N)


def bayesian_blocks(t, x=None, sigma=None,
                    fitness='events', **kwargs):
    """Bayesian Blocks Implementation
    This is a flexible implementation of the Bayesian Blocks algorithm
    described in Scargle 2012 [1]_
    Parameters
    ----------
    t : array_like
        data times (one dimensional, length N)
    x : array_like (optional)
        data values
    sigma : array_like or float (optional)
        data errors
    fitness : str or object
        the fitness function to use.
        If a string, the following options are supported:
        - 'events' : binned or unbinned event data
            extra arguments are `p0`, which gives the false alarm probability
            to compute the prior, or `gamma` which gives the slope of the
            prior on the number of bins.
        - 'regular_events' : non-overlapping events measured at multiples
            of a fundamental tick rate, `dt`, which must be specified as an
            additional argument.  The prior can be specified through `gamma`,
            which gives the slope of the prior on the number of bins.
        - 'measures' : fitness for a measured sequence with Gaussian errors
            The prior can be specified using `gamma`, which gives the slope
            of the prior on the number of bins.  If `gamma` is not specified,
            then a simulation-derived prior will be used.
        Alternatively, the fitness can be a user-specified object of
        type derived from the FitnessFunc class.
    Returns
    -------
    edges : ndarray
        array containing the (N+1) bin edges
    Examples
    --------
    Event data:
        t = np.random.normal(size=100)
        bins = bayesian_blocks(t, fitness='events', p0=0.01)
    Event data with repeats:
        t = np.random.normal(size=100)
        t[80:] = t[:20]
        bins = bayesian_blocks(t, fitness='events', p0=0.01)
    Regular event data:
        dt = 0.01
        t = dt * np.arange(1000)
        x = np.zeros(len(t))
        x[np.random.randint(0, len(t), int(len(t) / 10))] = 1
        bins = bayesian_blocks(t, x, fitness='regular_events', dt=dt, gamma=0.9)
    Measured point data with errors:
        t = 100 * np.random.random(100)
        x = np.exp(-0.5 * (t - 50) ** 2)
        sigma = 0.1
        x_obs = np.random.normal(x, sigma)
        bins = bayesian_blocks(t, x=x_obs, fitness='measures')
    References
    ----------
    .. [1] Scargle, J `et al.` (2012)
           http://adsabs.harvard.edu/abs/2012arXiv1207.5578S
    See Also
    --------
    astroML.plotting.hist : histogram plotting function which can make use
                            of bayesian blocks.
    """
    # validate array input
    t = np.asarray(t, dtype=float)
    if x is not None:
        x = np.asarray(x)
    if sigma is not None:
        sigma = np.asarray(sigma)

    # verify the fitness function
    if fitness == 'events':
        if x is not None and np.any(x % 1 > 0):
            raise ValueError("x must be integer counts for fitness='events'")
        fitfunc = Events(**kwargs)
    elif fitness == 'regular_events':
        if x is not None and (np.any(x % 1 > 0) or np.any(x > 1)):
            raise ValueError("x must be 0 or 1 for fitness='regular_events'")
        fitfunc = RegularEvents(**kwargs)
    elif fitness == 'measures':
        if x is None:
            raise ValueError("x must be specified for fitness='measures'")
        fitfunc = PointMeasures(**kwargs)
    else:
        if not (hasattr(fitness, 'args') and
                hasattr(fitness, 'fitness') and
                hasattr(fitness, 'prior')):
            raise ValueError("fitness not understood")
        fitfunc = fitness

    # find unique values of t
    t = np.array(t, dtype=float)
    assert t.ndim == 1
    unq_t, unq_ind, unq_inv = np.unique(t, return_index=True,
                                        return_inverse=True)

    # if x is not specified, x will be counts at each time
    if x is None:
        if sigma is not None:
            raise ValueError("If sigma is specified, x must be specified")

        if len(unq_t) == len(t):
            x = np.ones_like(t)
        else:
            x = np.bincount(unq_inv)

        t = unq_t
        sigma = 1

    # if x is specified, then we need to sort t and x together
    else:
        x = np.asarray(x)

        if len(t) != len(x):
            raise ValueError("Size of t and x does not match")

        if len(unq_t) != len(t):
            raise ValueError("Repeated values in t not supported when "
                             "x is specified")
        t = unq_t
        x = x[unq_ind]

    # verify the given sigma value
    N = t.size
    if sigma is not None:
        sigma = np.asarray(sigma)
        if sigma.shape not in [(), (1,), (N,)]:
            raise ValueError('sigma does not match the shape of x')
    else:
        sigma = 1

    # validate the input
    fitfunc.validate_input(t, x, sigma)

    # compute values needed for computation, below
    if 'a_k' in fitfunc.args:
        ak_raw = np.ones_like(x) / sigma / sigma
    if 'b_k' in fitfunc.args:
        bk_raw = x / sigma / sigma
    if 'c_k' in fitfunc.args:
        ck_raw = x * x / sigma / sigma

    # create length-(N + 1) array of cell edges
    edges = np.concatenate([t[:1],
                            0.5 * (t[1:] + t[:-1]),
                            t[-1:]])
    block_length = t[-1] - edges

    # arrays to store the best configuration
    best = np.zeros(N, dtype=float)
    last = np.zeros(N, dtype=int)

    #-----------------------------------------------------------------
    # Start with first data cell; add one cell at each iteration
    #-----------------------------------------------------------------
    for R in range(N):
        # Compute fit_vec : fitness of putative last block (end at R)
        kwds = {}

        # T_k: width/duration of each block
        if 'T_k' in fitfunc.args:
            kwds['T_k'] = block_length[:R + 1] - block_length[R + 1]

        # N_k: number of elements in each block
        if 'N_k' in fitfunc.args:
            kwds['N_k'] = np.cumsum(x[:R + 1][::-1])[::-1]

        # a_k: eq. 31
        if 'a_k' in fitfunc.args:
            kwds['a_k'] = 0.5 * np.cumsum(ak_raw[:R + 1][::-1])[::-1]

        # b_k: eq. 32
        if 'b_k' in fitfunc.args:
            kwds['b_k'] = - np.cumsum(bk_raw[:R + 1][::-1])[::-1]

        # c_k: eq. 33
        if 'c_k' in fitfunc.args:
            kwds['c_k'] = 0.5 * np.cumsum(ck_raw[:R + 1][::-1])[::-1]

        # evaluate fitness function
        fit_vec = fitfunc.fitness(**kwds)

        A_R = fit_vec - fitfunc.prior(R + 1, N)
        A_R[1:] += best[:R]

        i_max = np.argmax(A_R)
        last[R] = i_max
        best[R] = A_R[i_max]

    #-----------------------------------------------------------------
    # Now find changepoints by iteratively peeling off the last block
    #-----------------------------------------------------------------
    change_points = np.zeros(N, dtype=int)
    i_cp = N
    ind = N
    while True:
        i_cp -= 1
        change_points[i_cp] = ind
        if ind == 0:
            break
        ind = last[ind - 1]
    change_points = change_points[i_cp:]
    bins = edges[change_points]
    bins[0] = -np.inf
    bins[-1] = np.inf

    return bins

if __name__ == '__main__':
    t = [23, 4, 1, 2, 3, 4, 3, 3, 34, 54, 34, 54, 34, 43, 45, 22, 34, 25]
    bins = bayesian_blocks(t, fitness='events', p0=0.01)
    print(bins)


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2020/12/28 15:04
Desc:
'''

import os
import re
import json
import webbrowser

pattern_headdata=re.compile("var wqadd_headdata = \[\];",re.MULTILINE | re.DOTALL)
pattern_taildata=re.compile("var wqadd_taildata = \[\];",re.MULTILINE | re.DOTALL)
pattern_headfunc=re.compile("var wqadd_headfunc = \[\];",re.MULTILINE | re.DOTALL)
pattern_tailfunc=re.compile("var wqadd_tailfunc = \[\];",re.MULTILINE | re.DOTALL)
pattern_gpslist=re.compile(r"(?<=var marker, lineArr = ).*?(?=;$)",re.MULTILINE | re.DOTALL)
pattern_markerlist=re.compile(r"(?<=var marker_list = ).*?(?=;$)",re.MULTILINE | re.DOTALL)
pattern_automarker=re.compile(r"(?<=var auto_marker = ).*?(?=;$)",re.MULTILINE | re.DOTALL)
pattern_speed=re.compile(r"(?<=var _speed = ).*?(?=;$)",re.MULTILINE | re.DOTALL)

class Trajectory():
    def __init__(self, gps_list=[], marker_list=[], auto_marker=True):
        self.js=get_main_js()
        self.gps_str = json.dumps(gps_list)
        self.marker_str = json.dumps(marker_list)
        self.auto_marker = str(auto_marker).lower()
        self.head_data = ''
        self.tail_data = ''
        self.head_func = ''
        self.tail_func = ''
        self.html_name = os.path.abspath('gps_trajectory_playback.html')

    def add_jsdata(self,data_js='',position='head'):
        assert position in ('head','tail')
        if 'head'==position:
            self.head_data = str(data_js)
        else:
            self.tail_data = str(data_js)

    def add_jsfunc(self,func_js='',position='tail'):
        assert position in ('head','tail')
        if 'head'==position:
            self.head_func = str(func_js)
        else:
            self.tail_func = str(func_js)

    def building(self,html_name=None,web_open=False,speed=200):
        _out1 = re.sub(pattern_gpslist, self.gps_str, self.js)
        _out2 = re.sub(pattern_markerlist, self.marker_str, _out1)
        _out3 = re.sub(pattern_headdata, self.head_data, _out2)
        _out4 = re.sub(pattern_taildata, self.tail_data, _out3)
        _out5 = re.sub(pattern_headfunc, self.head_func, _out4)
        _out6 = re.sub(pattern_tailfunc, self.tail_func, _out5)
        _out = re.sub(pattern_automarker, self.auto_marker, _out6)
        _out = re.sub(pattern_speed, str(speed), _out)
        if html_name:
            self.html_name = os.path.abspath(str(html_name)+'.html')
        dump_data(_data=_out,file_name=self.html_name)
        if web_open:
            webbrowser.open(url='file://' + self.html_name, new=0, autoraise=True)

def get_main_js():
    main_js = '''
<!doctype html>
    <html>
    <head>
        <meta charset="utf-8">
        <meta http-equiv="X-UA-Compatible" content="IE=edge">
        <meta name="viewport" content="initial-scale=1.0, user-scalable=no, width=device-width">
        <title>轨迹回放</title>
        <link rel="stylesheet" href="https://a.amap.com/jsapi_demos/static/demo-center/css/demo-center.css"/>
        <script src="https://cache.amap.com/lbs/static/es5.min.js"></script>
        <script type="text/javascript" src="https://webapi.amap.com/maps?v=1.4.15&key=d0b0f942970a97389a81d82e505642bf&&plugin=AMap.Scale,AMap.OverView,AMap.ToolBar"></script>
        <style>
            html, body, #container {
                height: 100%;
                width: 100%;
            }
            .input-card .btn{
                margin-right: 1.2rem;
                width: 9rem;
            }
            .input-card .btn:last-child{
                margin-right: 0;
            }
        </style>
    </head>
    <body>
        <div id="container"></div>
        <div class="input-card">
            <h4>轨迹回放控制</h4>
            <div class="input-item">
                <input type="button" class="btn" value="开始动画" id="start" onclick="startAnimation()"/>
                <input type="button" class="btn" value="暂停动画" id="pause" onclick="pauseAnimation()"/>
            </div>
            <div class="input-item">
                <input type="button" class="btn" value="继续动画" id="resume" onclick="resumeAnimation()"/>
                <input type="button" class="btn" value="停止动画" id="stop" onclick="stopAnimation()"/>
            </div>
        </div>
        <script>
            var wqadd_headdata = [];
            var marker, lineArr = [[116.478935,39.997761],[116.478939,39.997825],[116.478912,39.998549],[116.478912,39.998549],[116.478998,39.998555],[116.478998,39.998555],[116.479282,39.99856],[116.479658,39.998528],[116.480151,39.998453],[116.480784,39.998302],[116.480784,39.998302],[116.481149,39.998184],[116.481573,39.997997],[116.481863,39.997846],[116.482072,39.997718],[116.482362,39.997718],[116.483633,39.998935],[116.48367,39.998968],[116.484648,39.999861]];
            if (lineArr.length<1)
            {
                marker_visible=false
            }
            else 
            {
                marker_visible=true
            }
            // 
            var map = new AMap.Map("container", {
                resizeEnable: true,
                center: lineArr[0],
                zoom: 17
            });
            var scale = new AMap.Scale({
                visible: true
            });
            var toolBar = new AMap.ToolBar({
                visible: true
            });
            var overView = new AMap.OverView({
                visible: true
            });
            map.addControl(scale);
            map.addControl(toolBar);
            map.addControl(overView);

            marker = new AMap.Marker({
                map: map,
                position: lineArr[0],
                icon: "https://webapi.amap.com/images/car.png",
                offset: new AMap.Pixel(-26, -13),
                autoRotation: true,
                angle:-90,
                visible:marker_visible,
            });
            // 绘制轨迹
            var polyline = new AMap.Polyline({
                map: map,
                path: lineArr,
                showDir:true,
                strokeColor: "#28F",  //线颜色
                // strokeOpacity: 1,     //线透明度
                strokeWeight: 6,      //线宽
                // strokeStyle: "solid"  //线样式
            });

            var passedPolyline = new AMap.Polyline({
                map: map,
                // path: lineArr,
                strokeColor: "#AF5",  //线颜色
                // strokeOpacity: 1,     //线透明度
                strokeWeight: 6,      //线宽
                // strokeStyle: "solid"  //线样式
            });

            //新增标记点
            var marker_list = [[116.478935,39.997761],[116.484648,39.999861]];
            var auto_marker = true;
            for (var i = 0; i < marker_list.length; i++) {
                var _curData = marker_list[i];
                if (_curData.length>=3)
                {
                    cont_text=(i+1).toString()+": "+_curData[2].toString();
                }
                else
                {   if (auto_marker)
                    {
                        cont_text=(i+1).toString();
                    }
                    else
                    {
                        cont_text='';
                    }     
                }
                var marker_i = new AMap.Marker({position: new AMap.LngLat(_curData[0],_curData[1]),label:{content: cont_text}});
                    map.add(marker_i);
            };
            marker.on('moving', function (e) {
                passedPolyline.setPath(e.passedPath);
            });
            map.setFitView();

            var wqadd_taildata = [];
            var wqadd_headfunc = [];

            function startAnimation () {
                var _speed = 200;
                marker.moveAlong(lineArr, _speed);
            }
            function pauseAnimation () {
                marker.pauseMove();
            }
            function resumeAnimation () {
                marker.resumeMove();
            }
            function stopAnimation () {
                marker.stopMove();
            }

            var wqadd_tailfunc = [];
        </script>
    </body>
</html>
'''
    return main_js

def dump_data(_data, file_name):
    with open(file_name, 'w') as f:
        data_html = f.write(_data)
    print("The output file is downloaded to {}".format(file_name))


if __name__ == '__main__':
    gps_list=[[116.478935,39.997761],[116.478939,39.997825],[116.478912,39.998549],[116.478912,39.998549],[116.478998,39.998555],[116.478998,39.998555],[116.479282,39.99856],[116.479658,39.998528],[116.480151,39.998453],[116.480784,39.998302],[116.480784,39.998302],[116.481149,39.998184],[116.481573,39.997997],[116.481863,39.997846],[116.482072,39.997718],[116.482362,39.997718],[116.483633,39.998935],[116.48367,39.998968],[116.484648,39.999861]]
    marker_list=[[116.478935,39.997761],[116.484648,39.999861]]

    traj = Trajectory(gps_list=gps_list,marker_list=marker_list,auto_marker=False)
    traj.building(web_open=True,speed=20000)






# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-30 14:42
Desc:
    标准评分卡(lr)
    log(odds)=w*x
    score = self.score_offset + self.score_factor * np.log(odds)
'''

from pynovice.score_card.src.score_card import ScoreCardModel
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
import pandas as pd
import numpy as np

class LRModel(ScoreCardModel):
    '''
        用于预测某一条预处理过的特征向量得分的方法
    :return:
        score_label: 最终预测的评分
        score_feature: 各个特征的得分字典
    '''

    def __init__(self,**kwargs):
        super().__init__()
        self.model_ = None
        self.woe_score = None
        self.kwargs = kwargs

    def train(self,train_feature, train_label, test_feature=pd.DataFrame, test_label=pd.DataFrame,eval=True,eval_plot=False):
        if (test_feature.empty or test_label.empty):
            train_feature, test_feature, train_label, test_label = \
                train_test_split(train_feature, train_label, test_size=0.2, random_state=0)

        self.model_ = LogisticRegression(**self.kwargs)
        self.model_.fit(train_feature, train_label)
        if eval:
            self.woe_score = self.get_card_info(train_feature)
            # 模型评估
            print("training eval:")
            df_pre, _ = self.predict(train_feature)
            auc, _ = self.get_auc(df_pre, train_label)
            ks, _ = self.get_ks(df_pre, train_label)
            lift, _ = self.get_lift(df_pre, train_label)
            print("auc={}, ks={}, lift={}".format(auc, ks['gap'].values[0], lift))

            print('testing eval:')
            df_pre, _ = self.predict(test_feature)
            auc, _ = self.get_auc(df_pre, test_label)
            ks, _ = self.get_ks(df_pre, test_label)
            lift, _ = self.get_lift(df_pre, test_label)
            print("auc={}, ks={}, lift={}".format(auc, ks['gap'].values[0], lift))

        if eval_plot:
            df_pre, _ = self.predict(test_feature)
            self.plot_roc(df_pre, test_label, pre_target=1, save_path='.')
            self.plot_ks(df_pre, test_label, pre_target=1, save_path='.')
            self.plot_lift(df_pre, test_label, pre_target=1, save_path='.')

    def predict(self,x):
        '''
        :param x: list/datafame
        :return: array([proba]),array([score_label])
        '''
        proba = self.model_.predict_proba(x)[:,1]
        score_label = np.array([self.probability_to_score(bad_prob=proba_i) for proba_i in proba])
        return proba,score_label

    def predict_detail(self,x):
        '''

        :param x: list/datafame
        :return: base_score,array([field_score]),array([final_score])
        '''
        coef_ = self.model_.coef_[0]
        intercept_ = self.model_.intercept_[0]

        base_score = round(self.score_offset + self.score_factor * intercept_,2)  # 基础分
        field_score = (np.array(x)*coef_*self.score_factor).round(decimals=2)  # 特征的分值
        final_score = (base_score + field_score.sum(axis=1)).astype(int)  #综合评分
        return base_score, field_score, final_score

    def get_card_info(self,df_field):
        '''
        计算评分卡模型详细分数信息
        :param df_field:
        :return:
        '''
        _out = []
        base_score = None
        for i,field in enumerate(df_field.columns):
            df_idx = df_field[field].drop_duplicates().index
            df_eval = df_field.loc[df_idx,:]
            base_score, field_score, _ = self.predict_detail(x=df_eval)
            # 变量评分
            df_iter = pd.DataFrame(df_eval[field].values,columns=['value'])
            df_iter['score'] = field_score[:,i]
            df_iter.insert(loc=0, column='fields', value=field)
            _out.append(df_iter.sort_values('value'))
        df_card = pd.concat(_out, axis=0)
        df_card.loc['__'] = ['_base',None, base_score]
        df_card = df_card.sort_values('fields').reset_index(drop=True)
        return df_card

if __name__ == '__main__':
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 6, 8, 3, 2, 1, 5, 7],[10, 2, 3, 42, 534, 5, 53, 6, 83, 3, 42, 21, 25, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field1', 'field2', 'label']
    df_fields, df_label = df[['field1','field2']], df['label']
    lr = LRModel(class_weight='balanced')
    lr.train(df_fields,df_label,eval=True)
    print(lr.predict_detail([[2,8]]))
    print(lr.predict_detail(df_fields.head()))

    print(lr.predict([[2,8]]))
    print(lr.predict(df_fields.head()))




# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-30 14:42
Desc:
    基于违约率的评分模型
    score = self.score_offset + self.score_factor * np.log(odds)
'''

from pynovice.score_card.src.score_card import ScoreCardModel
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

class BaseModel(ScoreCardModel):
    '''
        用于预测某一条预处理过的特征向量得分的方法
    :return:
        score_label: 最终预测的评分
    '''

    def __init__(self,bad_target=1,eps=1e-4):
        super().__init__()
        self.bad_target = bad_target
        self.eps = eps
        # _out
        self.base_logodds = None
        self.fields_logodds = None

    def train(self,train_feature, train_label, test_feature=pd.DataFrame, test_label=pd.DataFrame, eval=True,eval_plot=False):
        if (test_feature.empty or test_label.empty):
            train_feature, test_feature, train_label, test_label = \
                train_test_split(train_feature, train_label, test_size=0.2, random_state=0)

        train_feature.reset_index(drop=True, inplace=True)
        train_label.reset_index(drop=True, inplace=True)
        base_bad_rate = 1.00 * (train_label == self.bad_target).sum() / train_label.size
        self.base_logodds = max(base_bad_rate, self.eps) / max(1 - base_bad_rate, self.eps)

        fields_logodds = []
        for iname in train_feature.columns:
            df_concat = pd.concat([train_feature[iname], train_label], axis=1)
            _log_odds = df_concat.groupby(iname).agg(
                lambda x: 1.00 * np.log(
                    max(sum(x == self.bad_target), self.eps) / max(sum(x != self.bad_target), self.eps)))
            odds_dict = dict(zip(_log_odds.index, _log_odds.values[:, 0]))
            fields_logodds.append(odds_dict)
        self.fields_logodds = fields_logodds

        if eval:
            self.woe_score = self.get_card_info(train_feature)
            # 模型评估
            print("training eval:")
            df_pre, _ = self.predict(train_feature)
            auc, _ = self.get_auc(df_pre, train_label)
            ks, _ = self.get_ks(df_pre, train_label)
            lift, _ = self.get_lift(df_pre, train_label)
            print("auc={}, ks={}, lift={}".format(auc, ks['gap'].values[0], lift))

            print('testing eval:')
            df_pre, _ = self.predict(test_feature)
            auc, _ = self.get_auc(df_pre, test_label)
            ks, _ = self.get_ks(df_pre, test_label)
            lift, _ = self.get_lift(df_pre, test_label)
            print("auc={}, ks={}, lift={}".format(auc, ks['gap'].values[0], lift))

        if eval_plot:
            df_pre, _ = self.predict(test_feature)
            self.plot_roc(df_pre, test_label, pre_target=1, save_path='.')
            self.plot_ks(df_pre, test_label, pre_target=1, save_path='.')
            self.plot_lift(df_pre, test_label, pre_target=1, save_path='.')

    def predict(self,x):
        '''
        :param x: list/datafame
        :return: array([proba]),array([score_label])
        '''
        _, _, final_score = self.predict_detail(x)
        proba = np.array([self.score_to_probability(score_i) for score_i in final_score])
        return proba, final_score

    def predict_detail(self,x):
        '''

        :param x: list/datafame
        :return: base_score,array([field_score]),array([final_score])
        '''
        if isinstance(x,(pd.DataFrame,pd.Series)):
            x = x.values.tolist()

        intercept_ = 0
        logodds = np.array([list(self.fields_logodds[i].get(v,self.base_logodds) for i,v in enumerate(iterm)) for iterm in x])
        base_score = round(self.score_offset + self.score_factor * intercept_,2)  # 基础分
        field_score = (logodds*self.score_factor).round(decimals=2)  # 特征的分值
        final_score = (base_score + field_score.sum(axis=1)).astype(int)  #综合评分
        return base_score, field_score, final_score

    def get_card_info(self,df_field):
        '''
        计算评分卡模型详细分数信息
        :param df_field:
        :return:
        '''
        _out = []
        base_score = None
        for i,field in enumerate(df_field.columns):
            df_idx = df_field[field].drop_duplicates().index
            df_eval = df_field.loc[df_idx,:]
            base_score, field_score, _ = self.predict_detail(x=df_eval)
            # 变量评分
            df_iter = pd.DataFrame(df_eval[field].values,columns=['value'])
            df_iter['score'] = field_score[:,i]
            df_iter.insert(loc=0, column='fields', value=field)
            _out.append(df_iter.sort_values('value'))
        df_card = pd.concat(_out, axis=0)
        df_card.loc['__'] = ['_base',None, base_score]
        df_card = df_card.sort_values('fields').reset_index(drop=True)
        return df_card


if __name__ == '__main__':
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 6, 8, 3, 2, 1, 5, 7],[10, 2, 3, 42, 534, 5, 53, 6, 83, 3, 42, 21, 25, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field1', 'field2', 'label']
    df_fields, df_label = df[['field1','field2']], df['label']

    base = BaseModel()
    base.train(df_fields,df_label,eval=True)
    print(base.predict_detail([[2,10]]))
    print(base.predict([[2,10]]))

    print(base.predict_detail(df_fields))
    print(base.predict(df_fields))



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-06-29 11:14
Desc:
监控评分卡开发样本和现有样本分数分布的差异程度，关注样本人群的稳定性。
1. Population Stability Index (PSI)：群体稳定性指标，用来衡量分数分布的变化
2. Characteristic Stability Index (CSI)：特征稳定性指标，用来衡量特征层面的变化
'''
import pickle
import pandas as pd
import numpy as np
from pynovice.score_card.src.data_binning import DataBinning

class StabilityIndex:
    def __init__(self):
        '''
        计算模型分的稳定性psi：
            需要加载bins&bins_statistic_dict，可计算calc_expect_bins_statistic函数
        计算入模特征的稳定性csi：
            需要加载bins&bins_statistic_dict和bins_statistic_dict，可计算calc_expect_bins_statistic和load_expect_bins_score函数
        或者直接加载保存到本地的数据load_bins_statistic()/load_bins_score()
        '''
        self.bins_statistic_dict = {}  # 各箱的count数量
        self.bins_score_dict = {}  # 各箱的对应贡献分
        self.bins=[]

    def calc_expect_bins_statistic(self,expect_array,bins=10.0,bins_func='frequence_blocks'):
        '''
          对expect数组进行分箱统计
        :param expect_array:
        :param bins: bins_list或者bins数
        :param bins_func: bins为整数时使用bins_func进行分箱
        :return:
        '''
        df = pd.Series(expect_array).dropna()
        if isinstance(bins,list):
            self.bins=bins
            df_bins = pd.cut(df, bins=bins)
        else:
            binning = DataBinning(box_num=bins, _func=bins_func)
            df_bins = binning.fit_transform(df_x=df)
            self.bins = binning.bins
        self.bins_statistic_dict = df_bins.value_counts().to_dict()

    def calc_expect_bins_score(self,bins_score_dict):
        '''

        :param bins_score_dict:  各分bin对应的模型分值
        :return:
        '''
        self.bins_score_dict = bins_score_dict

    def load_bins_statistic(self,file_name):
        with open(file_name, "rb") as f:
            self.bins, self.bins_statistic_dict = pickle.load(f)

    def load_bins_score(self,file_name):
        with open(file_name, "rb") as f:
            self.bins_score_dict = pickle.load(f)

    def dump_bins_statistic(self,file_name):
        # json not suppert Interval type
        bins_info = [self.bins, self.bins_statistic_dict]
        with open(file_name, "wb") as f:
            pickle.dump(bins_info, f)

    def get_psi(self,actual_array,eps=1e-4):
        '''
        psi = (实际占比-预期占比)/LN(实际占比/预期占比)
        :param actual_array:
        :param eps:
        :return:
        '''
        df_expect = pd.DataFrame.from_dict(self.bins_statistic_dict,orient='index',columns=['expected_samples'])
        df_actual = pd.DataFrame(pd.cut(actual_array,self.bins).value_counts(),columns=['actual_samples'])
        df_out = pd.concat([df_expect, df_actual], axis=1)
        df_out['expected_rate'] = df_out['expected_samples']/max(eps,df_out['expected_samples'].sum())
        df_out['actual_rate'] = df_out['actual_samples'] / max(eps,df_out['actual_samples'].sum())
        df_out['rate_change'] = df_out['actual_rate']-df_out['expected_rate']
        df_out['ln(actual/expect)'] = (df_out['actual_rate'] / df_out['expected_rate']).map(
            lambda x: np.log(max(eps, x)))
        df_out['psi'] = df_out['rate_change']*df_out['ln(actual/expect)']
        columns_round = ['expected_rate', 'actual_rate', 'rate_change', 'ln(actual/expect)', 'psi']
        df_out[columns_round] = df_out[columns_round].round(4)
        return df_out

    def get_csi(self,actual_array,eps=1e-4):
        '''
         csi = (实际占比-预期占比)*每箱得分
        :param actual_array:
        :param eps:
        :return:
        '''
        df_psi = self.get_psi(actual_array)
        df_score_attribute = pd.DataFrame.from_dict(self.bins_score_dict, orient='index',
                                                    columns=['score_attribute']).round(4)
        df_out = pd.concat([df_psi, df_score_attribute], axis=1)
        df_out['csi'] = (df_out['rate_change'] * df_out['score_attribute']).round(4)
        return df_out

if __name__ == '__main__':
    expected_array, actual_array=[20,13,14,1,16,1,1,1,1],[20,15,17,15,16,1,1,1,1]

    # 1.psi:
    stab_index = StabilityIndex()
    stab_index.calc_expect_bins_statistic(expect_array=expected_array,bins=3)
    #stab_index.dump_bins_statistic('./model_score_bins_statistic.pkl')
    aa = stab_index.get_psi(actual_array)
    print(aa)
    bins_score_dict=dict(zip(aa.index,[20.0,30.0,40.0]))
    # 2.csi:
    stab_index = StabilityIndex()
    stab_index.calc_expect_bins_statistic(expect_array=expected_array, bins=3)
    stab_index.calc_expect_bins_score(bins_score_dict)
    bb = stab_index.get_csi(actual_array)
    print(bb)



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-28 11:17
Desc:
    best_ks 分箱 分割
    将特征值值进行从小到大的排序。
    计算出KS最大的那个值(或者是bin size最大的)，即为切点，记为D。然后把数据切分成两部分。
    重复步骤2，进行递归，D左右的数据进一步切割。直到KS的箱体数达到我们的预设阈值即可。

    这里实现的是 ks最大和bin size 最大进行交替寻找
'''
import pandas as pd
import numpy as np

def calc_ks(df_field, df_label):
    '''
        计算ks值
    :param df_field:
    :param df_label:
    :return:
    '''
    # 好坏的个数
    label_statue = df_label.value_counts()
    good_cnt, bad_cnt = label_statue[0], label_statue[1]
    # 每个段不同的好坏数量
    data_cro = pd.crosstab(df_field, df_label)
    data_cro.sort_index(inplace=True)
    # 计算不同段的好坏百分比
    data_cro[0] = data_cro[0] / good_cnt
    data_cro[1] = data_cro[1] / bad_cnt
    # 数值进行累加
    data_cro_cum = data_cro.cumsum()
    # 计算不同段的好坏比差别
    ks_list = abs(data_cro_cum[1] - data_cro_cum[0])
    return ks_list

def ks_bin(df_field, df_label, limit):
    '''
        找到满足条件的最大的ks值划分位置
    :param df_field:
    :param df_label:
    :param limit:
    :return:ks_i 满足条件的分割点，最大的ks值
    '''
    if np.unique(df_label).size<2:
        return None,-np.inf
    ks_list = calc_ks(df_field, df_label)

    ks_sorted = ks_list.sort_values(ascending=False, inplace=False)
    for ks_i, ks in ks_sorted.items():
        data_1 = df_field[df_field <= ks_i]
        data_2 = df_field[df_field > ks_i]
        if len(data_1) >= limit and len(data_2) >= limit:
            break

    if len(ks_list)>1:
        ks_index = ks_list.index.astype(float).to_list()
        ks_index.sort()
        pos_i = ks_index.index(ks_i)
        if pos_i==len(ks_index)-1:
            ks_i=sum(ks_index[-2:])/2.0
        else:
            ks_i = sum(ks_index[pos_i:pos_i+2]) / 2.0
    return ks_i,ks

def get_best_split(ks_measure, _focus='ks'):
    ks_measure = np.array(ks_measure)  # 分割点，ks值，binsize
    ks_i = ks_measure[:,0]
    ind = 2 if _focus == 'bin_size' else 1
    idx = ks_measure[:,ind].argmax()
    return ks_i[idx]

def ks_blocks(df_field, df_label,box_num=5 ,bad_target=1):
    '''
    best_ks 分箱：找到最大ks分割点后，之后循环从数量最多的bins中找ks分隔点
    :param df_field:
    :param df_label:
    :param box_num: 分箱数量
    :param bad_target:
    :return:
    '''
    df_field.reset_index(drop=True, inplace=True)
    df_label.reset_index(drop=True, inplace=True)
    df_label.apply(lambda x:1 if x==bad_target else 0)
    df_concat = pd.concat([df_field, df_label], axis=1)
    df_concat.columns = ['field', 'label']
    df_concat.sort_values(by='field', ascending=True, inplace=True)

    limit_ = df_concat.shape[0] / 20  # 总体的5%(bins要大于5%)

    ks_i,ks = ks_bin(df_concat['field'],df_concat['label'], limit_)
    df_set={ks_i:(df_concat,ks,len(df_concat))}

    ks_list = list()
    for i in range(box_num - 1):
        ks_measure = [[ks_i, v[1], v[2]] for ks_i, v in df_set.items()] # 分割点，ks值，binsize
        _focus = 'ks' if i%2==0 else 'bin_size' ## ks值和bin_size交叉寻找最大切分点
        ks_i = get_best_split(ks_measure,_focus)  # 找到最佳分割点
        if not ks_i:
            continue
        ks_list.append(ks_i)
        df_tmp,ks,bin_size = df_set.pop(ks_i)
        df_data1 = df_tmp[df_tmp['field'] > ks_i]
        df_data2 = df_tmp[df_tmp['field'] <= ks_i]
        ks_i_1, ks_1 = ks_bin(df_data1['field'], df_data1['label'], limit_)
        ks_i_2, ks_2 = ks_bin(df_data2['field'], df_data2['label'], limit_)
        df_set.update({ks_i_1:(df_data1,ks_1,len(df_data1))}) #{分割点：(data,ks,bin_size)}
        df_set.update({ks_i_2: (df_data1, ks_2,len(df_data2))})

    ks_list.sort()
    blocks = np.concatenate([[-np.inf],ks_list,[np.inf]])

    return blocks

if __name__ == '__main__':
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 3, 3, 3, 2, 1, 5, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    #df = pd.DataFrame([[3, 4, 0, 1, 2, 2, 2, 5, 7, 4, 7, 7, 7, 6], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field', 'label']
    df_field, df_label = df['field'], df['label']
    aa = ks_blocks(df_field, df_label,box_num=5, bad_target=1)
    print(aa)



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-05-04 15:34
Desc:
    特征生成（预处理，转换，编码）
'''
from pynovice.score_card.src.category_coder import CateCoder
from pynovice.score_card.src.data_binning import DataBinning
from pynovice.score_card.src.data_woe import WeightOfEvidence
import pandas as pd
from pynovice.util import progress_bar
import logging

class FeatureGenerator:
    def __init__(self,auto_fill_missing=True,missing=None):
        self.cate_coder_dict = {}
        self.binning_dict = {}
        self.woe_dict = {}
        self.woe_defalut = {}
        self.fields = None #特征字段
        self.category_fields = None
        self.digital_fields = None
        if missing:
            self.missing = missing
            self.auto_fill_missing = False
        else:
            self.missing = -999
            self.auto_fill_missing = auto_fill_missing
        self.binning_woe_info = None
        self.cate_fields_info = None

    def fit_transform(self, df_fields, df_label, category_fields=None, default_boxs=5, boxs_dict={},bins_dict={}):
        '''
            预处理+类别转换+分箱+woe编码
        :param df_fields: 训练集pandas.DataFrame
        :param df_label: Series
        :param category_fields: df_fields中类别特征字段，list
        :param default_boxs: 默认分箱数量
        :param boxs_dict: 自定义分箱数量
        :param bins_dict: 自定义分箱阈值，{field:[]}
        :return:
        '''
        df_fields = df_fields.copy()
        df_fields=df_fields.reset_index(drop=True)
        df_label = df_label.reset_index(drop=True)
        self.fields = df_fields.columns.to_list()
        if category_fields is None:
            self.category_fields = list()
        else:
            self.category_fields = category_fields
        self.digital_fields = list(set(self.fields)-set(self.category_fields))

        # 1.预处理
        df_fields = self.preprocessing(df_fields)

        # 2.类别特征转换
        for field in category_fields:
            cate_coder = CateCoder()
            df_fields[field] = cate_coder.fit_transform(df_fields[field], df_label)
            self.cate_coder_dict.update({field:cate_coder})

        # 3.分箱&woe编码
        for field in self.fields:
            _bins = bins_dict.get(field)
            # 自定义分箱
            if bool(_bins) & isinstance(_bins,list):
                box_num = len(_bins)
                binning = DataBinning(box_num=box_num)
                binning.bins = _bins
                df_field = pd.cut(df_fields[field],bins=_bins)
            # 程序分箱
            else:
                box_num = boxs_dict.get(field,default_boxs)
                binning = DataBinning(box_num=box_num, _func='tree_blocks')
                df_field = binning.fit_transform(df_x=df_fields[field], df_y=df_label)
            self.binning_dict.update({field:binning})

            woe = WeightOfEvidence()
            df_field = woe.fit_transform(df_field, df_label)
            self.woe_dict.update({field: woe})
            self.woe_defalut.update({field:df_field.mode()[0]})## 众数

            df_fields.loc[:, field] = df_field

        self.binning_woe_info = self.get_binning_woe()
        self.cate_fields_info = self.get_cate_info()
        return df_fields

    def transform(self,fields_jsons):
        if isinstance(fields_jsons,pd.DataFrame):
            return self.transform_for_dataframe(fields_jsons)
        if not isinstance(fields_jsons,list):
            fields_jsons = [fields_jsons]
        features=[]
        len_data = len(fields_jsons)
        for i,fields_json in enumerate(fields_jsons):
            progress_bar(i,len_data)
            feature = []
            for field_name in self.fields:
                # 主逻辑
                # 默认值
                if self.auto_fill_missing:
                    _missing = self.woe_defalut.get(field_name)
                else:
                    _missing = self.missing
                # 取数据
                field_value = fields_json.get(field_name)
                if not field_value:
                    feature.append(_missing)
                    continue
                # 类别编码
                cate_coder = self.cate_coder_dict.get(field_name)
                if cate_coder:
                    field_value = cate_coder.transform(field_value)
                if not isinstance(field_value,(float,int)):
                    logging.error("{} is not a number type[index:{}]".format(field_name,i))  # 数值类型
                    feature.append(_missing)
                    continue
                # data分箱
                binning = self.binning_dict.get(field_name)
                if binning:
                    field_value = binning.transform(field_value)
                # woe编码
                woe = self.woe_dict.get(field_name)
                if woe:
                    field_value = woe.transform(field_value)
                    if not field_value:
                        field_value = _missing
                else:
                    field_value = _missing
                feature.append(field_value)
            features.append(feature)
        return features

    def transform_for_dataframe(self,df_fields):
        df_in = self.preprocessing(df_fields)
        feature_columns=[]
        for field_name in self.fields:
            feature_columns.append(field_name)
            # 默认值
            if self.auto_fill_missing:
                _missing = self.woe_defalut.get(field_name)
            else:
                _missing = self.missing
            # 取数据
            if field_name not in df_in.columns:
                df_in[field_name]=_missing
                continue
            # 类别编码
            cate_coder = self.cate_coder_dict.get(field_name)
            if cate_coder:
                df_in[field_name] = df_in[field_name].apply(lambda x:cate_coder.transform(x))
            # data分箱
            binning = self.binning_dict.get(field_name)
            if binning:
                df_in[field_name] = pd.cut(df_in[field_name], bins=binning.bins)
            # woe编码
            woe = self.woe_dict.get(field_name)
            if woe:
                df_in[field_name] = df_in[field_name].apply(lambda x: woe.transform(x)).fillna(_missing)
            else:
                df_in[field_name] = _missing
        features = df_in[feature_columns]
        return features

    def preprocessing(self, df_fields):
        _columns = list(set(df_fields.columns)&set(self.digital_fields))
        if _columns:
            df_fields[_columns] = df_fields[_columns].fillna(self.missing).astype(float)
        return df_fields

    def get_binning_woe(self):
        field_woe = [woe.woe_info for woe in self.woe_dict.values()]
        df_woe = pd.concat(field_woe,axis=0)
        return df_woe

    def get_cate_info(self):
        cate_fields_info = {field:cate.cate_dict for field,cate in self.cate_coder_dict.items()}
        return cate_fields_info

if __name__ == '__main__':
    # dump
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 6, 8, 3, 2, 1, 5, 7], [10, 2, 3, 42, 534, 5, 53, 6, 83, 3, 42, 21, 25, 7],
                       [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field1', 'field2', 'label']
    df_fields, df_label = df[['field1', 'field2']], df['label']

    feature_generator = FeatureGenerator()
    feature_generator.fit_transform(df_fields,df_label,category_fields=['field1'])
    print(feature_generator.cate_fields_info)
    print(feature_generator.binning_woe_info)

    # file_name = '../data/feature_generater.pkl'
    # pickle.dump(feature_generator, open(file_name, "wb"),protocol=3)
    #
    # # load
    # feature_generator= pickle.load(open(file_name, "rb"))
    fields_json = [{'field1':'4a', 'field2':'5a', 'aa':9},{'field1':4, 'field2':5, 'aa':9}]
    df_fields = pd.DataFrame(fields_json)
    feature = feature_generator.transform(df_fields)
    print(feature)


""" This defines some common colors.
"""

from __future__ import division

import numpy as np

from pynovice.score_card.src.card_explanations.utils._colorconv import lab2rgb, lch2lab

try:
    import matplotlib
    from matplotlib.colors import LinearSegmentedColormap

    def lch2rgb(x):
        return lab2rgb(lch2lab([[x]]))[0][0]

    # define our colors using Lch
    # note that we intentionally vary the lightness during interpolation so as to better
    # enable the eye to see patterns (since patterns are most easily recognized through
    # lightness variability)
    blue_lch = [54., 70., 4.6588]
    l_mid = 40.
    red_lch = [54., 90., 0.35470565 + 2* np.pi]
    gray_lch = [55., 0., 0.]
    blue_rgb = lch2rgb(blue_lch)
    red_rgb = lch2rgb(red_lch)
    gray_rgb = lch2rgb(gray_lch)
    white_rgb = np.array([1.,1.,1.])
    
    
    light_blue_rgb = np.array([127., 196, 252])/255
    light_red_rgb = np.array([255., 127, 167])/255

    # define a perceptually uniform color scale using the Lch color space
    reds = []
    greens = []
    blues = []
    alphas = []
    nsteps = 100
    l_vals = list(np.linspace(blue_lch[0], l_mid, nsteps//2)) + list(np.linspace(l_mid, red_lch[0], nsteps//2))
    c_vals = np.linspace(blue_lch[1], red_lch[1], nsteps)
    h_vals = np.linspace(blue_lch[2], red_lch[2], nsteps)
    for pos,l,c,h in zip(np.linspace(0, 1, nsteps), l_vals, c_vals, h_vals):
        lch = [l, c, h]
        rgb = lch2rgb(lch)
        reds.append((pos, rgb[0], rgb[0]))
        greens.append((pos, rgb[1], rgb[1]))
        blues.append((pos, rgb[2], rgb[2]))
        alphas.append((pos, 1.0, 1.0))

    red_blue = LinearSegmentedColormap('red_blue', {
        "red": reds,
        "green": greens,
        "blue": blues,
        "alpha": alphas
    })
    red_blue.set_bad(gray_rgb, 1.0)
    red_blue.set_over(gray_rgb, 1.0)
    red_blue.set_under(gray_rgb, 1.0) # "under" is incorrectly used instead of "bad" in the scatter plot

    red_blue_no_bounds = LinearSegmentedColormap('red_blue_no_bounds', {
        "red": reds,
        "green": greens,
        "blue": blues,
        "alpha": alphas
    })

    red_blue_transparent = LinearSegmentedColormap('red_blue_no_bounds', {
        "red": reds,
        "green": greens,
        "blue": blues,
        "alpha": [(a[0], 0.5, 0.5) for a in alphas]
    })

    # define a circular version of the color scale for categorical coloring
    reds = []
    greens = []
    blues = []
    alphas = []
    nsteps = 100
    c_vals = np.linspace(blue_lch[1], red_lch[1], nsteps)
    h_vals = np.linspace(blue_lch[2], red_lch[2], nsteps)
    for pos,c,h in zip(np.linspace(0, 0.5, nsteps), c_vals, h_vals):
        lch = [blue_lch[0], c, h]
        rgb = lch2rgb(lch)
        reds.append((pos, rgb[0], rgb[0]))
        greens.append((pos, rgb[1], rgb[1]))
        blues.append((pos, rgb[2], rgb[2]))
        alphas.append((pos, 1.0, 1.0))
    c_vals = np.linspace(red_lch[1], blue_lch[1], nsteps)
    h_vals = np.linspace(red_lch[2] - 2 * np.pi, blue_lch[2], nsteps)
    for pos,c,h in zip(np.linspace(0.5, 1, nsteps), c_vals, h_vals):
        lch = [blue_lch[0], c, h]
        rgb = lch2rgb(lch)
        reds.append((pos, rgb[0], rgb[0]))
        greens.append((pos, rgb[1], rgb[1]))
        blues.append((pos, rgb[2], rgb[2]))
        alphas.append((pos, 1.0, 1.0))

    red_blue_circle = LinearSegmentedColormap('red_blue_circle', {
        "red": reds,
        "green": greens,
        "blue": blues,
        "alpha": alphas
    })

    colors = []
    for l in np.linspace(1, 0, 100):
        colors.append((30./255, 136./255, 229./255,l))
    for l in np.linspace(0, 1, 100):
        colors.append((255./255, 13./255, 87./255,l))
    red_transparent_blue = LinearSegmentedColormap.from_list("red_transparent_blue", colors)

    colors = []
    for l in np.linspace(0, 1, 100):
        colors.append((30./255, 136./255, 229./255,l))
    transparent_blue = LinearSegmentedColormap.from_list("transparent_blue", colors)

    colors = []
    for l in np.linspace(0, 1, 100):
        colors.append((255./255, 13./255, 87./255,l))
    transparent_red = LinearSegmentedColormap.from_list("transparent_red", colors)

    old_blue_rgb = np.array([30, 136, 229]) / 255
    old_red_rgb = np.array([255, 13, 87]) / 255

    colors = []
    for alpha in np.linspace(1, 0, 100):
        c = blue_rgb * alpha + (1 - alpha) * white_rgb
        colors.append(c)
    for alpha in np.linspace(0, 1, 100):
        c = red_rgb * alpha + (1 - alpha) * white_rgb
        colors.append(c)
    red_white_blue = LinearSegmentedColormap.from_list("red_transparent_blue", colors)


except ImportError:
    pass

#default_colors = ["#1E88E5", "#ff0d57", "#13B755", "#7C52FF", "#FFC000", "#00AEEF"]

#blue_rgba = np.array([0.11764705882352941, 0.5333333333333333, 0.8980392156862745, 1.0])
# blue_rgba = np.array([30, 136, 229, 255]) / 255
# blue_rgb = np.array([30, 136, 229]) / 255
# red_rgb = np.array([255, 13, 87]) / 255

# default_blue_colors = []
# tmp = blue_rgba.copy()
# for i in range(10):
#     default_blue_colors.append(tmp.copy())
#     if tmp[-1] > 0.1:
#         tmp[-1] *= 0.7



# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/8/19 16:58
Desc:
'''
import sys
import re
import copy


def safe_isinstance(obj, class_path_str):
    """
    Acts as a safe version of isinstance without having to explicitly
    import packages which may not exist in the users environment.

    Checks if obj is an instance of type specified by class_path_str.

    Parameters
    ----------
    obj: Any
        Some object you want to test against
    class_path_str: str or list
        A string or list of strings specifying full class paths
        Example: `sklearn.ensemble.RandomForestRegressor`

    Returns
    --------
    bool: True if isinstance is true and the package exists, False otherwise
    """
    if isinstance(class_path_str, str):
        class_path_strs = [class_path_str]
    elif isinstance(class_path_str, list) or isinstance(class_path_str, tuple):
        class_path_strs = class_path_str
    else:
        class_path_strs = ['']

    # try each module path in order
    for class_path_str in class_path_strs:
        if "." not in class_path_str:
            raise ValueError("class_path_str must be a string or list of strings specifying a full \
                module path to a class. Eg, 'sklearn.ensemble.RandomForestRegressor'")

        # Splits on last occurence of "."
        module_name, class_name = class_path_str.rsplit(".", 1)

        # here we don't check further if the model is not imported, since we shouldn't have
        # an object of that types passed to us if the model the type is from has never been
        # imported. (and we don't want to import lots of new modules for no reason)
        if module_name not in sys.modules:
            continue

        module = sys.modules[module_name]

        # Get class
        _class = getattr(module, class_name, None)

        if _class is None:
            continue

        if isinstance(obj, _class):
            return True

    return False

def format_value(s, format_str):
    """ Strips trailing zeros and uses a unicode minus sign.
    """

    if not issubclass(type(s), str):
        s = format_str % s
    s = re.sub(r'\.?0+$', '', s)
    if s[0] == "-":
        # s = u"\u2212" + s[1:]
        s = "-" + s[1:]
    return s


def convert_color(color):
    try:
        color = pl.get_cmap(color)
    except:
        pass

    if color == "shap_red":
        color = colors.red_rgb
    elif color == "shap_blue":
        color = colors.blue_rgb

    return color

def convert_ordering(ordering, shap_values):
    if issubclass(type(ordering), OpChain):
        ordering = ordering.apply(Explanation(shap_values))
    if issubclass(type(ordering), Explanation):
        if "argsort" in [op["name"] for op in ordering.op_history]:
            ordering = ordering.values
        else:
            ordering = ordering.argsort.flip.values
    return ordering


def merge_nodes(values, partition_tree):
    """ This merges the two clustered leaf nodes with the smallest total value.
    """
    M = partition_tree.shape[0] + 1

    ptind = 0
    min_val = np.inf
    for i in range(partition_tree.shape[0]):
        ind1 = int(partition_tree[i, 0])
        ind2 = int(partition_tree[i, 1])
        if ind1 < M and ind2 < M:
            val = np.abs(values[ind1]) + np.abs(values[ind2])
            if val < min_val:
                min_val = val
                ptind = i
                # print("ptind", ptind, min_val)

    ind1 = int(partition_tree[ptind, 0])
    ind2 = int(partition_tree[ptind, 1])
    if ind1 > ind2:
        tmp = ind1
        ind1 = ind2
        ind2 = tmp

    partition_tree_new = partition_tree.copy()
    for i in range(partition_tree_new.shape[0]):
        i0 = int(partition_tree_new[i, 0])
        i1 = int(partition_tree_new[i, 1])
        if i0 == ind2:
            partition_tree_new[i, 0] = ind1
        elif i0 > ind2:
            partition_tree_new[i, 0] -= 1
            if i0 == ptind + M:
                partition_tree_new[i, 0] = ind1
            elif i0 > ptind + M:
                partition_tree_new[i, 0] -= 1

        if i1 == ind2:
            partition_tree_new[i, 1] = ind1
        elif i1 > ind2:
            partition_tree_new[i, 1] -= 1
            if i1 == ptind + M:
                partition_tree_new[i, 1] = ind1
            elif i1 > ptind + M:
                partition_tree_new[i, 1] -= 1
    partition_tree_new = np.delete(partition_tree_new, ptind, axis=0)

    # update the counts to be correct
    fill_counts(partition_tree_new)

    return partition_tree_new, ind1, ind2


def get_sort_order(dist, clust_order, cluster_threshold, feature_order):
    """ Returns a sorted order of the values where we respect the clustering order when dist[i,j] < cluster_threshold
    """

    # feature_imp = np.abs(values)

    # if partition_tree is not None:
    #     new_tree = fill_internal_max_values(partition_tree, shap_values)
    #     clust_order = sort_inds(new_tree, np.abs(shap_values))
    clust_inds = np.argsort(clust_order)

    feature_order = feature_order.copy()  # order.apply(Explanation(shap_values))
    # print("feature_order", feature_order)
    for i in range(len(feature_order) - 1):
        ind1 = feature_order[i]
        next_ind = feature_order[i + 1]
        next_ind_pos = i + 1
        for j in range(i + 1, len(feature_order)):
            ind2 = feature_order[j]

            # if feature_imp[ind] >
            # if ind1 == 2:
            #     print(ind1, ind2, dist[ind1,ind2])
            if dist[ind1, ind2] <= cluster_threshold:

                # if ind1 == 2:
                #     print(clust_inds)
                #     print(ind1, ind2, next_ind, dist[ind1,ind2], clust_inds[ind2], clust_inds[next_ind])
                if dist[ind1, next_ind] > cluster_threshold or clust_inds[ind2] < clust_inds[next_ind]:
                    next_ind = ind2
                    next_ind_pos = j
                    # print("next_ind", next_ind)
                    # print("next_ind_pos", next_ind_pos)

        # insert the next_ind next
        for j in range(next_ind_pos, i + 1, -1):
            # print("j", j)
            feature_order[j] = feature_order[j - 1]
        feature_order[i + 1] = next_ind
        # print(feature_order)

    return feature_order


def sort_inds(partition_tree, leaf_values, pos=None, inds=None):
    if inds is None:
        inds = []

    if pos is None:
        partition_tree = fill_internal_max_values(partition_tree, leaf_values)
        pos = partition_tree.shape[0] - 1

    M = partition_tree.shape[0] + 1

    if pos < 0:
        inds.append(pos + M)
        return

    left = int(partition_tree[pos, 0]) - M
    right = int(partition_tree[pos, 1]) - M

    left_val = partition_tree[left, 3] if left >= 0 else leaf_values[left + M]
    right_val = partition_tree[right, 3] if right >= 0 else leaf_values[right + M]

    if left_val < right_val:
        tmp = right
        right = left
        left = tmp

    sort_inds(partition_tree, leaf_values, left, inds)
    sort_inds(partition_tree, leaf_values, right, inds)

    return inds


class OpChain():
    """ A way to represent a set of dot chained operations on an object without actually running them.
    """

    def __init__(self, root_name=""):
        self._ops = []
        self._root_name = root_name

    def apply(self, obj):
        """ Applies all our ops to the given object.
        """
        for o in self._ops:
            op, args, kwargs = o
            if args is not None:
                obj = getattr(obj, op)(*args, **kwargs)
            else:
                obj = getattr(obj, op)
        return obj

    def __call__(self, *args, **kwargs):
        """ Update the args for the previous operation.
        """
        new_self = OpChain(self._root_name)
        new_self._ops = copy.copy(self._ops)
        new_self._ops[-1][1] = args
        new_self._ops[-1][2] = kwargs
        return new_self

    def __getitem__(self, item):
        new_self = OpChain(self._root_name)
        new_self._ops = copy.copy(self._ops)
        new_self._ops.append(["__getitem__", [item], {}])
        return new_self

    def __getattr__(self, name):
        # Don't chain special attributes
        if name.startswith("__") and name.endswith("__"):
            return None
        new_self = OpChain(self._root_name)
        new_self._ops = copy.copy(self._ops)
        new_self._ops.append([name, None, None])
        return new_self

    def __repr__(self):
        out = self._root_name
        for o in self._ops:
            op, args, kwargs = o
            out += "."
            out += op
            if (args is not None and len(args) > 0) or (kwargs is not None and len(kwargs) > 0):
                out += "("
                if args is not None and len(args) > 0:
                    out += ", ".join([str(v) for v in args])
                if kwargs is not None and len(kwargs) > 0:
                    out += ", " + ", ".join([str(k) + "=" + str(kwargs[k]) for k in kwargs.keys()])
                out += ")"
        return out


# -*- coding: utf-8 -*-
"""
Created on Fri Dec 25 11:16:25 2020

@author: wangwenhao
"""

import pynovice.step_wise as mpmr
from sklearn.datasets import make_classification,make_regression
import pandas as pd


def get_X_y(data_type):
    if data_type == 'logistic':
        #含有信息的变量个数：4个
        #冗余变量个数：2。 冗余变量是有信息变量的线性组合
        #无用变量个数=10-4-2=4。

        #number of informative features = 4
        #number of redundant features = 2.redundant feature is linear combinations of the informative features
        #number of useless features = 10-4-2=4
        X, y = make_classification(n_samples=200,n_features=10,n_informative=4,n_redundant=2,shuffle=False,random_state=0,class_sep=2)
        X = pd.DataFrame(X,columns=['informative_1','informative_2','informative_3','informative_4','redundant_1','redundant_2','useless_1','useless_2','useless_3','useless_4']).sample(frac=1)
        y=pd.Series(y).loc[X.index]

    if data_type == 'linear':
        # 含有信息的变量个数：6个
        # 特征矩阵的秩：2个（说明含有信息量的6个变量中存在共线性）

        # number of informative features = 6
        # matrix rank = 2 (implying collinearity between six informative features)
        X, y = make_regression(n_samples=200,n_features=10,n_informative=6,effective_rank=2,shuffle=False,random_state=0)#
        X = pd.DataFrame(X,columns=['informative_1','informative_2','informative_3','informative_4','informative_5','informative_6','useless_1','useless_2','useless_3','useless_4']).sample(frac=1)
        y=pd.Series(y).loc[X.index]
    return X, y

def test_logit(X,y):
    lr  =  mpmr.Regression(X,y,measure='roc_auc',iter_num=20, max_related_level=2,observation_model='all')
    in_vars,dr = lr.fit()
    return in_vars,dr


if __name__ == '__main__':
    import warnings
    warnings._setoption('ignore')
    X_logit, y_logit = get_X_y('logistic')
    in_vars_logit,dr_logit = test_logit(X_logit,y_logit)

    print(in_vars_logit)



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-06-29 11:12
Desc:
'''

from ._waterfall import waterfall
from ._force import force
# from ._beeswarm import beeswarm
from ._beeswarm import summary_legacy as summary_plot



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-07-28 15:07
Desc: 数据分布画图
'''


from .anysis_distribution_plot import DistributionAnysis
from .echarts_base_plot import *


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/8/16 14:47
Desc:
'''

import numpy as np
import pandas as pd
from sklearn.metrics import explained_variance_score
from sklearn.metrics import r2_score
from sklearn.metrics import max_error
from sklearn.metrics import accuracy_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import balanced_accuracy_score
from sklearn.metrics import average_precision_score
from statsmodels.stats.outliers_influence import variance_inflation_factor
import logging


class _Tool():
    def KS(y_true, y_hat, sample_weight=None):
        if isinstance(y_true, np.ndarray):
            y_true = pd.Series(y_true)
        if sample_weight is None:
            sample_weight = pd.Series(np.ones_like(y_true), index=y_true.index)
        if isinstance(y_hat, np.ndarray):
            y_hat = pd.Series(y_hat, index=y_true.index)
        sample_weight.name = 'sample_weight'
        y_true.name = 'y'
        y_hat.name = 'score'
        df = pd.concat([y_hat, y_true, sample_weight], axis=1)
        df['y_mutli_w'] = df['y'] * df['sample_weight']
        total = df.groupby(['score'])['sample_weight'].sum()
        bad = df.groupby(['score'])['y_mutli_w'].sum()
        all_df = pd.DataFrame({'total': total, 'bad': bad})
        all_df['good'] = all_df['total'] - all_df['bad']
        all_df.reset_index(inplace=True)
        all_df = all_df.sort_values(by='score', ascending=False)
        all_df['badCumRate'] = all_df['bad'].cumsum() / all_df['bad'].sum()
        all_df['goodCumRate'] = all_df['good'].cumsum() / all_df['good'].sum()
        ks = all_df.apply(lambda x: x.goodCumRate - x.badCumRate, axis=1)
        return np.abs(ks).max()

    def vif(df):
        vif = pd.DataFrame()
        vif['features'] = df.columns
        if df.shape[1] > 1:
            vif['VIF Factor'] = [variance_inflation_factor(df.values, i) for i in range(df.shape[1])]
        else:
            vif['VIF Factor'] = 0
        vif = vif.sort_values('VIF Factor', ascending=False)
        return vif

    def make_logger(is_logger, logger_file):
        logger_level = logging.INFO if is_logger else logging.DEBUG
        logging.getLogger().setLevel(logger_level)
        fh = None
        if logger_file is not None:
            fh = logging.FileHandler(logger_file, mode='w')
            fh.setLevel(logging.DEBUG)
            formatter = logging.Formatter(
                '[%(name)s]-[%(filename)s-%(lineno)d]-[%(processName)s]-[%(asctime)s]-[%(levelname)s]: %(message)s')
            fh.setFormatter(formatter)
            logger.addHandler(fh)
        return logging, fh

SCORERS = dict(
    r2=r2_score,
    explained_variance_score=explained_variance_score,
    max_error=max_error,
    accuracy=accuracy_score,
    roc_auc=roc_auc_score,
    balanced_accuracy=balanced_accuracy_score,
    average_precision=average_precision_score,
    ks=_Tool.KS)


# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2020/9/16 17:34
Desc:
'''
import gmplot
import webbrowser
import os

# plot the locations on google map  真实经纬度信息
def plot_google_map(data, to_html_name,web_open=False):
    html_name = os.path.abspath(to_html_name)
    assert 'latitude' in data.columns
    assert 'longitude' in data.columns
    latitude_array = data['latitude'].values
    latitude_list = latitude_array.tolist()
    Longitude_array = data['longitude'].values
    longitude_list = Longitude_array.tolist()
    # Initialize the map to the first location in the list
    gmap = gmplot.GoogleMapPlotter(latitude_list[0], longitude_list[0], 10)
    # gmap.scatter(latitude_list, longitude_list, edge_width=10)
    gmap.heatmap(latitude_list, longitude_list)
    # Write the map in an HTML file
    # gmap.draw('Paths_map.html')
    gmap.draw(html_name)
    if web_open:
        webbrowser.open(url='file://' + html_name, new=0, autoraise=True)

if __name__ == '__main__':
    import pandas as pd
    df_raw = pd.DataFrame(data=[[-0.4302744, 35.5331974],
                                [-1.21167, 36.8973031],
                                [0.5217472, 35.255192],
                                [-2.2326433, 38.0769783],
                                [-1.0625246, 34.4734754]],
                          columns=['latitude', 'longitude']
                          )
    plot_google_map(df_raw,"google_apply_map.html",web_open=True)



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-22 16:58
Desc:
    1.数据分箱(离散化数据)
    代优化
'''
from collections import Sequence

import pandas as pd
import numpy as np
from pynovice.score_card.src.data_boxing import *

DISCRETIZATION_FUNCTION = {
    'frequence_blocks':
        lambda df_x, df_y, box_num: frequence_blocks(df_x, box_num),
    'distince_blocks':
        lambda df_x, df_y, box_num: distince_blocks(df_x, box_num),
    'kmeans_blocks':
        lambda df_x, df_y, box_num: kmeans_blocks(df_x, box_num),
    'bayesian_blocks':
        lambda df_x, df_y, box_num: bayesian_blocks(t=df_x, x=None, sigma=None, fitness='events'),
    'ks_blocks':
        lambda df_x, df_y, box_num: ks_blocks(df_x, df_y, box_num=box_num, bad_target=1),
    'chi_blocks':
        lambda df_x, df_y, box_num: chi_blocks(df_x, df_y, box_num=box_num, dfree=4, cf=0.1,max_iterations=100),
    'tree_blocks':
        lambda df_x, df_y, box_num: tree_blocks(df_x, df_y, box_num=box_num, criterion='entropy')
}

class DataBinning:
    """离散化连续数据
    Attributes:
         bins (Sequence): - 用于分段的列表,第一位为下限,最后一位为上限
    """

    def __init__(self, box_num=5, _func='tree_blocks'):
        self.box_num = box_num
        self.func = self.check_func(_func)
        # _out
        self.bins = None

    def fit(self, df_x, df_y=None):
        '''
        数据分箱
        :param x:
        :param y:
        :param _func: 分箱采用的方法，默认为二叉树分箱
        :return: 分箱切割点
        '''
        blocks = self.func(df_x,df_y,self.box_num)
        self.bins = blocks

    def transform(self, x):
        if not x:
            return [-np.inf,np.inf]
        s = pd.cut([x], bins=self.bins)[0]
        return s

    def fit_transform(self, df_x, df_y=None):
        self.fit(df_x, df_y)
        out = pd.cut(df_x, bins=self.bins)
        return out

    def check_func(self,_func):
        if isinstance(_func,str):
            _key = DISCRETIZATION_FUNCTION.keys()
            assert _func in _key, "func_list: {}".format(_key)
            _func = DISCRETIZATION_FUNCTION.get(_func)
        return _func

if __name__ == '__main__':
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 3, 3, 3, 2, 1, 5, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field', 'label']
    df_field, df_label = df['field'], df['label']

    binning = DataBinning(box_num=2, _func='tree_blocks')
    bb=binning.fit_transform(df_x=df['field'],df_y=df_label)
    cc= binning.transform(3)
    print(bb)
    print(binning.bins)

    print(cc)




# -*- coding: utf-8 -*-
# /usr/bin/env python

'''
Author: wenqiangw
Email: wqwangchn@163.com
Date: 2021/01/19 15:04
Desc:
'''

import os
import re
import json
import webbrowser

pattern_headdata=re.compile("var add_data = \[\];",re.MULTILINE | re.DOTALL)
pattern_tailfunc=re.compile("var add_func = \{\};",re.MULTILINE | re.DOTALL)

pattern_gpslist=re.compile(r"(?<=var lineArr = ).*?(?=;$)",re.MULTILINE | re.DOTALL)
pattern_triplist=re.compile(r"(?<=var tripList = ).*?(?=;$)",re.MULTILINE | re.DOTALL)
pattern_markerlist=re.compile(r"(?<=var markerSet = ).*?(?=;$)",re.MULTILINE | re.DOTALL)
pattern_showMarker=re.compile(r"(?<=var showMarker = ).*?(?=;$)",re.MULTILINE | re.DOTALL)

CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))
class Trajectory():
    def __init__(self, lnglat_arr=[], trip_list=[], show_marker=True):
        assert show_marker in (True,False)
        self.js=get_main_js()
        self.gps_str = json.dumps(lnglat_arr)
        self.trip_str = json.dumps(trip_list)
        self.show_marker = str(show_marker).lower()
        self.trip_length=len(trip_list)
        self.marker_set = []
        self.head_data = []
        self.tail_func = []
        self.html_name = os.path.abspath('gps_trajectory_playback.html')

    def add_marker(self,marker_arr=[],marker_img='default'):
        if marker_img == 'default':
            if self.trip_length>0:
                img = "https://a.amap.com/jsapi_demos/static/demo-center/icons/poi-marker-red.png"
            else:
                img = "https://a.amap.com/jsapi_demos/static/demo-center/icons/poi-marker-default.png"
        elif marker_img == 'blue':
            img = "https://a.amap.com/jsapi_demos/static/demo-center/icons/poi-marker-default.png"
        elif marker_img == 'red':
            img="https://a.amap.com/jsapi_demos/static/demo-center/icons/poi-marker-red.png"
        else:
            img=marker_img
        if len(marker_arr)>0:
            self.marker_set.append({
                "data":marker_arr,
                "img":img
            })

    def add_jsdata(self,data_js=''):
        self.head_data.add(str(data_js))

    def add_jsfunc(self,func_js=''):
        self.tail_func.add(str(func_js))

    def building(self,html_name=None,web_open=False):
        _out = re.sub(pattern_gpslist, self.gps_str, self.js)
        _out = re.sub(pattern_triplist, self.trip_str, _out)
        _out = re.sub(pattern_showMarker, self.show_marker, _out)

        if self.marker_set:
            _out = re.sub(pattern_markerlist, json.dumps(self.marker_set), _out)
        _out = re.sub(pattern_headdata, ''.join(self.head_data), _out)
        _out = re.sub(pattern_tailfunc, ''.join(self.tail_func), _out)
        out = _out
        if html_name:
            self.html_name = os.path.abspath(str(html_name)+'.html')
        dump_data(_data=out,file_name=self.html_name)
        if web_open:
            webbrowser.open(url='file://' + self.html_name, new=0, autoraise=True)

def get_main_js():
    filename= os.path.join(CURRENT_PATH,'pynovice_map.js')
    with open(filename,'r') as f:
        main_js = f.read()
    return main_js

def dump_data(_data, file_name):
    with open(file_name, 'w') as f:
        data_html = f.write(_data)
    print("The output file is downloaded to {}".format(file_name))


if __name__ == '__main__':
    gps_list=[[116.478935,39.997761],[116.478939,39.997825],[116.478912,39.998549],[116.478912,39.998549],[116.478998,39.998555],[116.478998,39.998555],[116.479282,39.99856],[116.479658,39.998528],[116.480151,39.998453],[116.480784,39.998302],[116.480784,39.998302],[116.481149,39.998184],[116.481573,39.997997],[116.481863,39.997846],[116.482072,39.997718],[116.482362,39.997718,'info:ttttt'],[116.483633,39.998935,'info:qqqqqqqq'],[116.48367,39.998968,'info:1234254325'],[116.484648,39.999861,'info:dsafaf']]
    marker_list=[3,5,10,15]

    traj = Trajectory(lnglat_arr=gps_list,trip_list=marker_list,show_marker=False)
    traj.add_marker(marker_arr=[[116.478912,39.998549,'akdfjkajfkjaksdf']])
    traj.add_marker(marker_arr=[[116.480784,39.998302]],marker_img='blue')
    traj.building(web_open=True)






# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-07-03 17:31
Desc: 规划中，，，，不可用
BMap
Bar
Boxplot
Funnel
Geo
HeatMap
line
bar
Map
Pie
Scatter
WordCloud

图形组件
组合组件


[Bar,Line]
Scatter
HeatMap
'''

import webbrowser
import os
import pandas as pd
from pyecharts.components import Table
from pyecharts import options as opts
from pyecharts.charts import Bar, Line, Scatter, Page, Tab, Grid, Timeline,HeatMap

CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))

class DistributionAnysis:
    def __init__(self):
        self.commit = "Analyze and plot the data "
        self.html_name = os.path.abspath('data_distribution.html')
        self.datas = []

    def add_data(self, _df, chart_type, title_name, subtitle_name='', xname='', yname='', global_opts={},
                 xaxis_opts={}, yaxis_opts={}):
        chart = self.check_chart(chart_type)
        chart.set_global_opts(title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
                            xaxis_opts=opts.AxisOpts(axislabel_opts=opts.LabelOpts(rotate=-15), name=xname),  # label旋转
                            yaxis_opts=opts.AxisOpts(axislabel_opts=opts.LabelOpts(formatter="{value}"), name=yname),
                            datazoom_opts=opts.DataZoomOpts(),  # 滑动条
                            toolbox_opts=opts.ToolboxOpts(),
                            **global_opts
                            )
        xaxis = list(map(str, _df.index))
        chart.add_xaxis(xaxis,**xaxis_opts)
        for field in _df.columns:
            chart.add_yaxis(series_name=str(field), yaxis_data=_df[str(field)].to_list(),**yaxis_opts)
        self.datas.append(chart)
        return chart

    def extend_data(self, _df, chart_type, xname='', yname='', xaxis_opts={}, yaxis_opts={},extend_opts={}):
        chart = self.check_chart(chart_type)
        xaxis = list(map(str, _df.index))
        chart.add_xaxis(xaxis,**xaxis_opts)
        for field in _df.columns:
            chart.add_yaxis(series_name=str(field), y_axis=_df[str(field)].to_list(),**yaxis_opts)
        chart.extend_axis(yaxis=opts.AxisOpts(yname=yname, position="right",splitline_opts=opts.SplitLineOpts(
                            is_show=True, linestyle_opts=opts.LineStyleOpts(opacity=1))),**extend_opts)
        raw_chart = self.datas[-1]
        new_chart = raw_chart.overlap(chart)
        self.datas[-1] = new_chart
        return chart

    def plot_summary(self,html_name=None,web_open=False,page_type='page',init_opts={},page_opts={}):
        if html_name:
            self.html_name = os.path.abspath(str(html_name)+'.html')

        assert page_type in ('grid', 'page', 'tab', 'timeline'), \
            '仅支持并行多图(grid),单页多图(page),选项卡多图(tab),时间线轮播多图(timeline)'

        if 'grid' == page_type:
            grid = Grid(**init_opts)  # 并行多图
            for idata in self.datas:
                grid.add(idata,**page_opts)
                grid.render(self.html_name)

        if 'page' == page_type:
            page = Page(**init_opts)  # 单页多图
            for idata in self.datas:
                page.add(idata,**page_opts)
                page.render(self.html_name)

        if 'tab' == page_type:
            tab = Tab(**init_opts)  # 选项卡多图
            for idata in self.datas:
                tab.add(idata,**page_opts)
                tab.render(self.html_name)

        if 'timeline' == page_type:
            timeline = Timeline(**init_opts)  # 时间线轮播多图
            for idata in self.datas:
                timeline.add(idata,**page_opts)
                timeline.render(self.html_name)

        if web_open:
            webbrowser.open(url='file://' + self.html_name, new=0, autoraise=True)

    def check_type(self,page_type):
        assert page_type in ('grid', 'page', 'tab', 'timeline'), \
            '仅支持并行多图(grid),单页多图(page),选项卡多图(tab),时间线轮播多图(timeline)'
        self.page_type=page_type

    def check_chart(self, chart_type):
        if isinstance(chart_type,str):
            assert chart_type in ('bar','line','scatter'), '仅支持bar|line|scatter字符定义，特殊chart需输入clas原型'
            if 'bar'==chart_type:
                return Bar()
            elif 'line'==chart_type:
                return Line()
            elif 'scatter'==chart_type:
                return Scatter()
        elif isinstance(chart_type,classmethod):
            return chart_type
        else:
            assert chart_type in ('bar','line','scatter'), '仅支持bar|line|scatter字符定义，特殊chart需输入clas原型'


class WqTab(Tab):
    def __init__(self, page_title: str = CurrentConfig.PAGE_TITLE, js_host: str = ""):
        super().__init__(page_title=page_title,js_host=js_host)


if __name__ == '__main__':
    df_raw = pd.DataFrame( data = [[-0.4302744, 35.5331974],
                                   [-1.21167  , 36.8973031],
                                   [ 0.5217472, 35.255192 ],
                                   [-2.2326433, 38.0769783],
                                   [-1.0625246, 34.4734754]],
                           columns=['latitude', 'longitude'])

    any = DistributionAnysis()
    any.add_data(_df=df_raw,title_name='aa1', subtitle_name='aa',chart_type='bar',yname='xname',xname='yname')
    any.add_data(_df=df_raw,title_name='bb1', subtitle_name='bb',chart_type='line')
    any.plot_summary(web_open=True,html_name=None,afdasf=23)

    a=Bar()
    a.add_js_funcs()
